#	zgdchart.py
#
#	Change History:
#	0.3.5
#
#	06 Jun 2000    Support stacked graphs, histograms
#    	18 May 2000    Urlencode encodes dates properly
#	12 Apr 2000    Fixes bugs introduced in 0.3.2
#		       Improved code readability	
#

import Acquisition
import AccessControl
import OFS
import urllib
import string
from Globals import HTMLFile, MessageDialog, Persistent
from Globals import package_home
from DocumentTemplate.DT_Util import *
from DocumentTemplate.DT_String import String
import os

# write to temporary files
from tempfile import mktemp, tempdir

#
# In this module, extend gdchart so
# that it can render itself
#
from gdchart import *

class ZGDChart(
	Acquisition.Implicit,
	Persistent,
	AccessControl.Role.RoleManager,
	OFS.SimpleItem.Item
	):
	'''
	A simple chart class which plots SQL data.
	The chart plots the first two columns of a SQL query result.

	This product uses gdchart0.94b, which in turn uses gd1.3
	The current version of gd 1.7.3 does not support gif creation due to the Unisys patent.

	Wrapper authored by <a href="mailto:teyc@bigfoot.com">Chui Tey</a>
	
	(c) Cognoware Australia P/L

	Software is provided without warranty.

	Credits:

	gdchart 	- Bruce Verdeaime   <A HREF="http://www.fred.net/brv/chart/">http://www.fred.net/brv/chart/</A>
	
	gdchart.py	- Michael Ray Steed <A HREF="http://members.xoom.com/_XMCM/msteed/software/gdchart.html"> http://members.xoom.com/_XMCM/msteed/software/gdchart.html </A>
	
	This product was created using mkproduct 0.6 and Zope Simple Skeleton v0.2,
	written by Sebastian L&uuml;hnsdorf (<a href="mailto:basti@beehive.de">basti@beehive.de</a>)

	For help, refer to the <A href="http://localhost/HelpSys/ObjectRef/ZGDChart/hs_main">Zope help system</A>
	'''

	name	=	'ZGDChart'
	id  	=	'ZGDChart'
	icon	=	'misc_/ZGDChart/Main'
	meta_type=	'ZGDChart'

	_properties = (
		{'id':'title','type':'string','mode':'w'},
		)

	manage_options=(
		{'label':'Properties','icon':icon,'action':'manage_main',  'target':'manage_main'},
		{'label':'Colors'    ,'icon':icon,'action':'manage_mainColor','target':'manage_main'},
		{'label':'Scale'     ,'icon':icon,'action':'manage_mainScale','target':'manage_main'},
		{'label':'Security'  ,'icon':icon,'action':'manage_access','target':'manage_main'},
		{'label':'View'      ,'icon':icon,'action':'index_html',   'target':'manage_main'},
		)

	__ac_permissions__=(
		('View management screens',('manage_tabs','manage_main')),
		('Change permissions',('manage_access',)),
		('Change ZGDChart',('manage_edit','manage_editColor'),('Manager',)),
		)

	#	There is a bit of duplication here
	#	Maybe comment this out in the future
	#	
	chart_types = ["Pie", "Line", "Area", "Bar", "Histogram", "Histogram_percent"]	
	'	List of chart types allowed '

	def home(self):
		"returns the directory where the product is located in"
		return package_home(globals())
	
	def version (self):
		"Return the version number of this product"		
		return open(self.home() + '/version.txt').read()
	
	def __init__(self,id='ZGDChart',title='ZGDChart', height=250, width=250, 
		SQL=None, chart_type='Pie', option=[]):
		
		self.id 	    	= id
		self.title  	   	= title
		self.height 	   	= height
		self.width  	   	= width
		self.SQL    	   	= SQL
		self.content_type	= 'image/gif'
		# self.tempfile		= self.home() + '/temp.gif'
			
		# Set the chart options
		self.bg_transparent	= 'bg_transparent' in option
		self.xaxis    		= 'xaxis' in option
		self.yaxis      	= 'yaxis' in option
		self.yaxis2		= 'yaxis2' in option
		self.label_line		= 'label_line' in option
		self.grid		= 'grid' in option
		self.border		= 'border' in option
		self.thumbnail		= 'thumbnail' in option
		self.chart_type 	= chart_type

		# Set the chart color
		self.bg_color		=0xFFFFFF
		self.title_color	=0x446688
		self.edge_color		=0x000000
		self.grid_color		=0x446688
		self.plot_color		=0x446688
		self.xlabel_color	=0x446688
		self.ylabel_color	=0x446688
		self.ylabel2_color	=0x446644
		self.xtitle_color	=0x446688
		self.ytitle_color	=0x446688
		self.ytitle2_color	=0x446644

		# ---------------------------------
		# Declare these variables because
		# we don't want to break compatibility
		# when these evetually get implemented
		
		# Set chart text
		self.charttitle 	= title
		self.xtitle		= ''
		self.ytitle		= ''
		self.ytitle2		= ''

		# Set chart text font sizes
		self.title_font		= GDC_LARGE
		self.xtitle_font	= GDC_SMALL
		self.ytitle_font	= GDC_SMALL
		self.ytitle2_font	= GDC_SMALL
		self.annotation_font= GDC_SMALL

		# Set miscellaneous items
		self.bar_width		=	10
		self.explode 		=	[0,0,0]
		self.ext_color		=	[0,0,0]
		self.ext_vol_color 	=	[0,0,0]
		self.generate_gif	=	1
		self.hard_graphheight	=	300
		self.hard_graphwidth	= 	400
		self.hard_size			=	0
		self.hard_xorig			=	0
		self.hard_yorig			=	0
		self.hlc_cap_width		=	0
		self.hlc_style			=	GDC_HLC_CLOSE_CONNECTED
		self.hold_img			=	0
		self.label_dist			=	20
		self.missing			= 	[]
		self.other_threshold		= 	10
		self.percent_labels		=	GDCPIE_PCT_NONE
		self.pie_color			= 	[0xAACCEE, 0x88AACC, 0x6688AA]
		self.thumbval			=	0
		self.zeroshelf			=	0
		
		# Set axis
		self.requested_ymax		=	None
		self.requested_ymin		=	None
		self.requested_yinterval	=	None
		self.xlabel_spacing		=	None
		self.ylabel_density		=	80
		self.xhist_interval		=	10
		self.ylabel_fmt			=	''
		self.ylabel2_fmt		=	''
		
		# Set three dimensional properties		
		self.threed_angle		=	0
		self.threed_depth		=	0
						
	manage_main=HTMLFile('manage/edit', globals())
	manage_mainColor = HTMLFile('manage/editColor', globals())
	manage_mainScale = HTMLFile('manage/editScale', globals())
    
	def manage_edit(self,title, xtitle, ytitle, width, height, SQL, chart_type, option=[], REQUEST=None):
		"""edit ZGDChart"""
		self.title      = title
		self.xtitle     = xtitle
		self.ytitle     = ytitle
		self.height     = height
		self.width      = width
		self.SQL        = SQL
		self.chart_type = chart_type
		self.bg_transparent = 'bg_transparent' in option
		self.label_line     = 'label_line' in option
		self.thumbnail 	    = 'thumbnail' in option
		self.border	    = 'border' in option
		self.grid	    = 'grid' in option
		self.xaxis      = 'xaxis' in option
		self.yaxis      = 'yaxis' in option
		self.yaxis2     = 'yaxis2' in option

		if REQUEST is not None:
			return MessageDialog(
				title='Edited',
				message='<strong>%s</strong> has been edited.' % self.id,
				action = './manage_main'
				)

	def manage_editColor(self, bg_color, edge_color, grid_color, plot_color,
						 title_color, xtitle_color, ytitle_color, ytitle2_color,
						 xlabel_color, ylabel_color, ylabel2_color,
						 REQUEST=None):
		""" edit ZGDChart colors
			all values to be passed as strings in hex, eg..
			manage_editColor(bg_color = "0xFFFFFF")
		"""
		self.bg_color   	= string.atoi(bg_color,   16)
		self.edge_color 	= string.atoi(edge_color, 16)
		self.grid_color 	= string.atoi(grid_color, 16)
		self.plot_color 	= string.atoi(plot_color, 16)
		self.title_color	= string.atoi(title_color, 16)
		self.xtitle_color 	= string.atoi(xtitle_color, 16)
		self.ytitle_color 	= string.atoi(ytitle_color, 16)
		self.ytitle2_color 	= string.atoi(ytitle2_color, 16)
		self.xlabel_color	= string.atoi(xlabel_color, 16)
		self.ylabel_color 	= string.atoi(ylabel_color, 16)
		self.ylabel2_color	= string.atoi(ylabel2_color, 16)
		
		if REQUEST is not None:
			return MessageDialog(
				title='Edited',
				message='<strong>%s</strong> has been edited.' % self.id,
				action = './manage_main'
				)

	def manage_editScale(self,
						 requested_ymax=None, requested_ymin=None, requested_yinterval=None,
						 xlabel_spacing=0, xhist_interval=10, ylabel_density=80,
						 ylabel_fmt='', ylabel2_fmt='', REQUEST=None):
		""" edit ZGDChart scale properties """		
		self.requested_ymax		=	requested_ymax		
		self.requested_ymin		=	requested_ymin
		self.requested_yinterval	=	requested_yinterval
		self.xlabel_spacing		=	xlabel_spacing
		self.ylabel_density		=	ylabel_density
		self.ylabel_fmt			=	ylabel_fmt
		self.ylabel2_fmt		=	ylabel2_fmt
		if xhist_interval>0:
			self.xhist_interval     =	int(xhist_interval)

		if REQUEST is not None:
			return MessageDialog(
				title='Edited',
				message='<strong>%s</strong> has been edited.' % self.id,
				action = './manage_main'
				)
		    
	def __call__(self, mapping={}, **kwargs):
		''' Just another version of __str__ '''
		# Not sure if this is going to block other inherited methods

		# combine mapping and kwargs into a single dictionary
		dict = {}
		for key,value in mapping.items():
			dict[key]=value
		for key,value in kwargs.items():
			dict[key]=value
		
		# roll our own urlencode because we want
		# / to be encoded
		# query = urllib.urlencode(kwargs)
		query = self.urlencode(dict)
		return '<IMG SRC="%s?%s" ALT="%s">' % (self.id, query, self.title)

	def urlencode(self, dict):
		l = []
		for k, v in dict.items():
			k = urllib.quote_plus(str(k))
			v = urllib.quote_plus(str(v),'')
			l.append(k + '=' + v)
		return string.join(l, '&')

	# following routines are inspired by Arpad Kiss's chart product
	def __str__(self,REQUEST=None,RESPONSE=None):
		"""
		Returns a string representation of the object <IMG SRC=xxx>
		"""
		return '<IMG SRC="%s" ALT="%s">' % (self.id, self.title)
	
	def index_html(self,REQUEST=None,RESPONSE=None):
		"""
		This is the default method. It calls the right renderer: Line,Column or Pie
		"""
		return eval("self.%s(REQUEST,RESPONSE)" % self.chart_type)

	def Line(self,REQUEST=None,RESPONSE=None):
		"""
		Renders dataset as a line chart
		"""
		return self._chart(GDC_LINE, RESPONSE)

	def Pie(self,REQUEST=None,RESPONSE=None):
		"""
		Renders dataset as a pie chart
		"""
		return self._chart(GDC_3DPIE, RESPONSE)

	def Bar(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a bar chart
		"""
		return self._chart(GDC_BAR, RESPONSE)

	def Area(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a bar chart
		"""
		return self._chart(GDC_AREA, RESPONSE)

	def Histogram(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a histogram chart
		"""
		return self._chart(GDC_BAR, RESPONSE)

	def Histogram_percent(self, REQUEST=None, RESPONSE=None):
		"""
		Renders dataset as a histogram chart
		with Y axis as percentage
		"""
		return self._chart(GDC_BAR, RESPONSE)
	
	def _chart(self, GDC_TYPE, RESPONSE=None):
		"""
		Renders dataset in the format requested
		"""
		self._setOption()
		self._setColor()
		self._setScale()
		#
		# see if histogram chart data is required
		#
		if self.chart_type=='Histogram' or self.chart_type=='Histogram_percent':
			chartdata = self._chartdata_hist()
		else:
			chartdata = self._chartdata2()
		#
		# get a temporary file name
		#
		tempdir = self.home()
		tempfile = mktemp()
		#
		# we need a syntax which supports
		# chart(GDC_TYPE, (self.width, self.height), tempfile, labels, data1 [, data2 [, ...]])
		#
		if len(chartdata)>2:
			apply(chart, list((GDC_TYPE, (self.width, self.height), tempfile)) + list(chartdata))
		else:
			labels, data = chartdata
			chart(GDC_TYPE, (self.width, self.height), tempfile, labels, data)
		#
		if RESPONSE: RESPONSE['content-type']=self.content_type
		#
		# read temp file, and delete it
		#
		filecontents = open(tempfile,'rb').read()
		os.remove(tempfile)
		return  filecontents
		
	def _setOption(self):
		option(title = self.title)
		option(xtitle = self.xtitle)
		option(ytitle = self.ytitle)
		option(bg_transparent = self.bg_transparent)
		option(grid = self.grid)
		option(border = self.border)
		option(label_line= self.label_line)
		option(thumbnail = self.thumbnail)
		option(xaxis = self.xaxis)
		option(yaxis = self.yaxis)
		option(yaxis2= self.yaxis2)
		option(set_color = (0xaabbcc, 0xffbbcc, 0xaabb55, 0x3388aa, 0xaa8833, 0x88aa33))	
		option(pie_color = (0xaabbcc, 0xffbbcc, 0xaabb55, 0x3388aa, 0xaa8833, 0x88aa33))	
	
	def _setColor(self):
		option(bg_color  	= self.bg_color)
		option(grid_color	= self.grid_color)
		option(plot_color	= self.plot_color)
		option(edge_color 	= self.edge_color)
		option(title_color	= self.title_color)
		option(xtitle_color 	= self.xtitle_color)
		option(ytitle_color 	= self.ytitle_color)
		option(ytitle2_color 	= self.ytitle2_color)
		option(xlabel_color 	= self.xlabel_color)
		option(ylabel_color 	= self.ylabel_color)
		option(ylabel2_color 	= self.ylabel2_color)

	def _setScale(self):
		#
		#  the only possible way to reset the scaling options
		#  to default values is to use a patched gdchart.pyd,
		#  and pass None as a parameter. self.requested_ymax and
		#  other attributes in the following block may be a None
		#
		try:
			option(requested_ymax =	self.requested_ymax)
		        option(requested_ymin =	self.requested_ymin)
		        option(requested_yinterval = self.requested_yinterval)
		        option(xlabel_spacing =	self.xlabel_spacing)
		        option(ylabel_density =	self.ylabel_density)
		except error:
			#
			#  probably haven't got a patched gdchart.pyd
			#  will try to run gdchart using previous scale settings
			#  it will be ugly but at least won't give an error page
			#
			print "gdchart error: probably not using the patched gdchart.pyd"
			pass;
			
		option(ylabel_fmt	=	self.ylabel_fmt )
		option(ylabel2_fmt	=	self.ylabel2_fmt)

	def _chartdata(self):
		"""
		return the first and second columns of sql query
    		results as a pair of tuples
    		deprecated: use _chartdata2
	        """
		rows = getattr(self, self.SQL)()
		labels= []
		data =  []
		for row in rows:
			# each row is a list of fields
 			fields = row
			#
			# Only insert rows with zero values if
			# it is not a pie chart

			#
			if self.chart_type!='Pie' or float(fields[1])>0:
				#
				#	Coerce the first field into a string
				#	and second field into a float
				#
        			labels.append( str  (fields[0]))
        			data.append  ( float(fields[1]))
    		#
    		# convert lists to tuples
    		# apparently, gdchart.chart() accepts lists
    		# as well, but I better check with
    		# M Steed first.
    		#
    		return tuple(labels), tuple(data)

	def _chartdata2(self):
		"""
		return all the columns of a sql query
		as tuples of tuples
		"""
		rows= getattr(self, self.SQL)()
		if len(rows)==0:
			return (),()
		#
		# each row is a list of fields
		#
		num_cols=len(rows[0])
		if num_cols<2:
			return (),()
		#
		# create an array of columns
		# each element of the array
		# holds a list
		#
		columns={}
		for col in range(num_cols):
			columns[col]=[]
		for row in rows:
			#
			# each row is a list of fields
			#
			fields = row
			#
			# if we are dealing with a Pie chart
			# and field[1] == 0, then do not append
			#
			if self.chart_type=='Pie' and float(fields[1])==0:
				pass
			else:
				for col in range(num_cols):
					columns[col].append(fields[col])
		#
		# cast the first column as string, the remaining
		# columns as floats
		#
		columns[0]=map(str,columns[0])
		for col in range(1,num_cols):
			columns[col]=map(float,columns[col])
		#
		# convert dictionary into a list
		#
		columns = columns.items()
		columns.sort()
		#
		# get rid of the key in each key,value pair
		#
		columns=map(lambda list:list[1], columns)
		#
		# return the results as a tuple of tuples
		#
		return tuple(map(tuple,columns))

	def _chartdata_hist(self):
		"""
		return the first column of a sql query
		as a histogram, grouped by self.xhist_interval
		"""
		rows= getattr(self, self.SQL)()
		columns={}
		interval=self.xhist_interval
		for row in rows:
			#
			# calculate the group which a value belongs
			# to
			#
			group = int(float(row[0])/interval+0.5)*interval
			try:
				columns[group]=columns[group]+1
			except KeyError:
				columns[group]=1
		#
		# convert dictionary into a list
		# of key, value pairs
		#
		columns = columns.items()
		columns.sort()
		#
		# define an anonymous function that would
		# convert each key, value pair into
		# list of keys, list of values
		#
		def g(x,y):
			return x[0]+[y[0]], x[1]+[y[1]]			
		columns=reduce(g, columns, ([],[]))
		#
		# convert columns from a tuple to a list
		# so that we can mutate each element
		#
		columns=list(columns)
		#
		# define an anonymous function that would
		# convert a group value into an interval
		# expression and apply the function to
		# every element in the first column
		#
		def h(x, interval=interval):
			return str(x) + '-' + str(x+interval)
		columns[0]=map(h, columns[0])
		#
		# if the frequency is in percentage
		# then change the values in the histogram
		#
		if self.chart_type == 'Histogram_percent':
			total = reduce(lambda x,y:x+y, columns[1])
			columns[1]=map(lambda x,total=total: float(x)/total, columns[1])			
		#
		# return the results as a tuple of tuples
		#
		return tuple(map(tuple,columns))

			
	def _labels(self):
		"""
		return the first column sql query results as a tuple
	        """
		rows = getattr(self, self.SQL)()
		tmp=()
		for row in rows:
		   tmp = tmp + (row[0],)
		return tmp
		
	def _data(self):
		rows = getattr(self, self.SQL)()
		tmp=()
		for row in rows:
		    tmp = tmp + (row[1],)
		return tmp

addZGDChartForm=HTMLFile('manage/add',globals())

def addZGDChart(self,id,title, height, width, SQL, chart_type, option=[], REQUEST=None):
	"""Create a ZGDChart and install it in its parent Folder
	The argument 'self' will be bound to the parent Folder.
	"""

	handle=ZGDChart(id, title, height, width, SQL, chart_type, option)
	handle.id=id
	handle.title=title
	self._setObject(id,handle)
	if REQUEST is not None:
		return self.manage_main(self, REQUEST)

